package javabot.modules;

import javabot.Module;
import javabot.Secretary;
import javabot.FileHandler;

import java.io.*;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Stack;
import java.util.StringTokenizer;

/**
 * Created by IntelliJ IDEA.<br>
 * User: CreoN<br>
 * Date: 2003-aug-28<br>
 * Time: 15:18:23<br>
 *
 * Copyright (C) 2003 Fredrik Rydn & Tobias Larsson
 * Full copyright statement can be found in JavaBot.java
 *
 */
public class FileServer extends Module
{
    // Member variables for module
    private int counter;

    // Member variables for DCC-session
    private InetAddress listeningAddress;
    private int startPortRange;
    private int endPortRange;
    private String rootFolder;

    // Member variables for DCC-chat-session
    private int dccChatPort;

    private ServerSocket dccChatListeningSocket;
    private Socket dccChatConnectionSocket;
    private BufferedReader dccChatFromClient;
    private DataOutputStream dccChatToClient;
    private boolean dccChatConnectionActive;

    // Member variables for DCC-file-session
    private int dccFilePort;

    private ServerSocket dccFileListeningSocket;
    private Socket dccFileConnectionSocket;
    private DataInputStream dccFileFromClient;
    private DataOutputStream dccFileToClient;
    private boolean dccFileConnectionActive;


    // Other member variables
    private Stack pathStack;
    private String sender;
    private Secretary secretary;

    public void runModule()
    {
        int nrOfMaxConnections = Integer.parseInt(getSecretary().getConfig().getProperty("DCCMaxConnections"));
        if(counter < nrOfMaxConnections)
        {
            new FileServer(getSender(), getSecretary());
            counter++;
        }
        else
        {
            getSecretary().sendRawIRCMessage("NOTICE " + getSender() + " :Too many clients connected");
        }
    }

    public FileServer()
    {
    }

    public FileServer(String sender, Secretary secretary)
    {
        this.sender = sender;
        this.secretary = secretary;
        pathStack = new Stack();

        // Reads listening hostname from config
        try
        {
            listeningAddress = InetAddress.getByName(secretary.getConfig().getProperty("DCCListeningHostname"));
        }
        catch (UnknownHostException e)
        {
            listeningAddress = null;
            FileHandler.writeErrorLog(e.getMessage(), e.getStackTrace());
        }

        // Reads portrange from config
        String portRange = secretary.getConfig().getProperty("DCCPortRange");
        startPortRange = Integer.parseInt(portRange.substring(0, portRange.indexOf("-")));
        endPortRange = Integer.parseInt(portRange.substring(portRange.indexOf("-") + 1, portRange.length()));

        // Reads root folder from config
        rootFolder = secretary.getConfig().getProperty("DCCRootFolder");


        dccChatConnectionActive = false;

        waitForConnection();
        handleDCCSession();
    }

    private void waitForConnection()
    {
        boolean portIsTaken = true;
        while(portIsTaken)
        {
            dccChatPort = (int)((Math.random() * (endPortRange - startPortRange)) + startPortRange);
            try
            {
                dccChatListeningSocket = new ServerSocket(dccChatPort, 0, listeningAddress);
                portIsTaken = false;
            }
            catch(IOException e)
            {
                FileHandler.writeErrorLog(e.getMessage(), e.getStackTrace());
            }
            catch(SecurityException e)
            {
                portIsTaken = true;
            }
        }

        secretary.sendRawIRCMessage("PRIVMSG " + this.sender + " :" + Secretary.CHAR_CTCP + "DCC CHAT chat " + this.secretary.getConnectedSocket().getLocalAddress().hashCode() + " " + dccChatPort + Secretary.CHAR_CTCP);

        while(!dccChatConnectionActive)
        {
            try
            {
                dccChatConnectionSocket = dccChatListeningSocket.accept();

                dccChatFromClient = new BufferedReader(new InputStreamReader(dccChatConnectionSocket.getInputStream()));
                dccChatToClient = new DataOutputStream(dccChatConnectionSocket.getOutputStream());
                dccChatConnectionActive = true;
            }
            catch(Exception e)
            {
                FileHandler.writeErrorLog(e.getMessage(), e.getStackTrace());
            }
        }
        counter--;

        try
        {
            dccChatListeningSocket.close();
        }
        catch (IOException e)
        {
            FileHandler.writeErrorLog(e.getMessage(), e.getStackTrace());
        }
    }

    private void handleDCCSession()
    {
        String textFromClient = "";

        try
        {
            dccChatToClient.writeBytes("Welcome to this FServe, run on " + Secretary.NAME + " " + Secretary.VERSION + ".\r\nAvailable commands: ls cd get\r\n");
        }
        catch(IOException e)
        {
            FileHandler.writeErrorLog(e.getMessage(), e.getStackTrace());
        }

        while(textFromClient != null)
        {
            try
            {
                textFromClient = dccChatFromClient.readLine();
                if(textFromClient != null)
                {
                    if(textFromClient.length() > 1 && textFromClient.substring(0, 2).equals("ls"))
                    {
                        listFiles();
                    }
                    else if(textFromClient.length() > 1 && textFromClient.substring(0, 2).equals("cd"))
                    {
                        changeDirectory(textFromClient);
                    }
                    else if(textFromClient.length() > 2 && textFromClient.substring(0, 3).equals("pwd"))
                    {
                        dccChatToClient.writeBytes(getPath(pathStack) + "\r\n");
                    }
                    else if(textFromClient.length() > 2 && textFromClient.substring(0, 3).equals("get"))
                    {
                        downloadFile(textFromClient);
                    }
                    else
                    {
                        dccChatToClient.writeBytes("Unknown command\r\n");
                    }
                }
            }
            catch(Exception e)
            {
                try
                {
                    dccChatToClient.writeBytes("An error occurred: " + e.getMessage() + "\r\n");
                }
                catch (IOException e1)
                {
                    FileHandler.writeErrorLog(e1.getMessage(), e1.getStackTrace());
                }
            }
        }
        dccChatConnectionActive = false;
    }

    private String getPath(Stack stack)
    {
        Stack tmpStack = (Stack)stack.clone();
        String path = "";

        while(!tmpStack.isEmpty())
        {
            path = (String)tmpStack.pop() + File.separator + path;
        }

        path = rootFolder + File.separator + path;

        return path;
    }

    private void listFiles() throws Exception
    {
        File currentFolder = new File(getPath(pathStack));
        dccChatToClient.writeBytes("Current folder: " + currentFolder.getPath() + "\r\n");
        File[] fileList = currentFolder.listFiles();

        for(int i = 0; i < fileList.length; i++)
        {
            String filename = fileList[i].getName();
            String stringToSend = "";
            if(fileList[i].isDirectory())
            {
                stringToSend = "<dir>" + fillWith(" ", 5) + filename + "\r\n";
            }
            else
            {
                stringToSend = fileList[i].length() + fillWith(" ", 10-Long.toString(fileList[i].length()).length()) + filename + "\r\n";
            }
            dccChatToClient.writeBytes(stringToSend);
        }
    }

    private void changeDirectory(String textFromClient) throws Exception
    {
        StringTokenizer pathStringTokenizer = new StringTokenizer(textFromClient);
        pathStringTokenizer.nextToken();
        String newPath = pathStringTokenizer.nextToken("\n");
        newPath = newPath.substring(1, newPath.length());

        Stack tmpStack = (Stack)pathStack.clone();

        if(newPath.charAt(0) == File.separatorChar) // empty the stack
        {
            while(!tmpStack.isEmpty())
            {
                tmpStack.pop();
            }
        }

        StringTokenizer pathTokenizer = new StringTokenizer(newPath, File.separator);
        boolean failed = false;
        while(pathTokenizer.hasMoreTokens())
        {
            String token = pathTokenizer.nextToken(File.separator);
            if(!changePath(token, tmpStack))
            {
                failed = true;
            }
        }

        if(!failed)
        {
            pathStack = tmpStack;
        }
        else
        {
            dccChatToClient.writeBytes("No such folder\r\n");
        }
    }

    private void downloadFile(String textFromClient) throws Exception
    {
        StringTokenizer pathStringTokenizer = new StringTokenizer(textFromClient);
        pathStringTokenizer.nextToken();
        String newPath = pathStringTokenizer.nextToken("\n");
        newPath = newPath.substring(1, newPath.length());

        Stack tmpStack = (Stack)pathStack.clone();

        if(newPath.charAt(0) == File.separatorChar) // empty the stack
        {
            while(!tmpStack.isEmpty())
            {
                tmpStack.pop();
            }
        }

        StringTokenizer pathTokenizer = new StringTokenizer(newPath, File.separator);
        boolean failed = false;
        String filename = "";
        while(pathTokenizer.hasMoreTokens())
        {
            String token = pathTokenizer.nextToken(File.separator);
            if(pathTokenizer.hasMoreTokens())
            {
                if(!changePath(token, tmpStack))
                {
                    failed = true;
                }
            }
            else
            {
                filename = token;
            }
        }

        if(!failed)
        {
            String pathToFile = getPath(tmpStack) + filename;
            if(new File(pathToFile).exists() && new File(pathToFile).isFile())
            {
                boolean portIsTaken = true;
                while(portIsTaken)
                {
                    dccFilePort = (int)((Math.random() * (endPortRange - startPortRange)) + startPortRange);
                    try
                    {
                        dccFileListeningSocket = new ServerSocket(dccFilePort, 0, listeningAddress);
                        portIsTaken = false;
                    }
                    catch (SecurityException e)
                    {
                        portIsTaken = true;
                    }
                }
                secretary.sendRawIRCMessage("PRIVMSG " + sender + " :" + Secretary.CHAR_CTCP + "DCC SEND " + pathToFile + " " + this.secretary.getConnectedSocket().getLocalAddress().hashCode() + " " + dccFilePort + " " + new File(pathToFile).length() + Secretary.CHAR_CTCP);
                while(!dccFileConnectionActive)
                {
                    try
                    {
                        dccFileConnectionSocket = dccFileListeningSocket.accept();

                        dccFileFromClient = new DataInputStream(dccFileConnectionSocket.getInputStream());
                        dccFileToClient = new DataOutputStream(dccFileConnectionSocket.getOutputStream());
                        dccFileConnectionActive = true;
                    }
                    catch(IOException e)
                    {
                        FileHandler.writeErrorLog(e.getMessage(), e.getStackTrace());
                    }
                }

                dccFileListeningSocket.close();

                boolean reachedEOFOrError = false;
                FileInputStream fis = new FileInputStream(new File(pathToFile));
                while(!reachedEOFOrError)
                {
                    byte[] bytesToSend = new byte[8192];
                    int returnValue = fis.read(bytesToSend);
                    if(returnValue == -1)
                    {
                        reachedEOFOrError = true;
                    }
                    else
                    {
                        dccFileToClient.write(bytesToSend);
                        byte[] bytesFromClient = new byte[4];
                        dccFileFromClient.read(bytesFromClient);
                        int returnFromClient = ((bytesFromClient[0] & 0xFF) << 24)
                              | ((bytesFromClient[1] & 0xFF) << 16)
                              | ((bytesFromClient[2] & 0xFF) << 8)
                              | (bytesFromClient[3] & 0xFF);
                        fis.mark(returnFromClient);
                    }
                }
                dccFileConnectionActive = false;
                fis.close();
            }
            else
            {
                dccChatToClient.writeBytes("No such file\r\n");
            }
        }
        else
        {
            dccChatToClient.writeBytes("No such folder\r\n");
        }
    }

    private boolean changePath(String path, Stack stack)
    {
        String currentPath = getPath(stack);

        if(path.equals(".."))
        {
            if(!stack.isEmpty())
            {
                stack.pop();
                return true;
            }
            else
            {
                return false;
            }
        }
        else
        {
            currentPath += File.separator + path;
            File tmpFile = new File(currentPath);
            if(tmpFile.exists() && tmpFile.isDirectory())
            {
                stack.push(path);
                return true;
            }
        }

        return false;
    }


    public String[] getCommands()
    {
        String[] commands = {"!dcc"};
        return commands;
    }

    public void gotResponse(String[] response, String[] thingsToRemember)
    {
    }

    public void timerExpired(String[] thingsToRemember)
    {
    }

    public String getHelpMessage()
    {
        return "Initiates a DCC chat session from the bot. Usage: !dcc";
    }

    private String fillWith(String string, int amount)
    {
        String returnString = "";
        for(int i = 0; i < amount; i++)
        {
            returnString += string;
        }

        return returnString;
    }

    public String getGroupName()
    {
        return "FileServer";
    }
}
